@@ -22,6 +22,8 @@ const Tweener = imports.ui.tweener;
 
 const FOCUS_ANIMATION_TIME = 0.15;
 
+const WINDOW_DND_SIZE = 256;
+
 const FRAME_COLOR = new Clutter.Color();
 FRAME_COLOR.from_pixel(0xffffffff);
 
@@ -29,6 +31,8 @@ const SCROLL_SCALE_AMOUNT = 100 / 5;
 
 const ZOOM_OVERLAY_FADE_TIME = 0.15;
 
+const DRAGGING_WINDOW_OPACITY = 100;
+
 // Define a layout scheme for small window counts. For larger
 // counts we fall back to an algorithm. We need more schemes here
 // unless we have a really good algorithm.
@@ -124,10 +128,12 @@ WindowClone.prototype = {
                            Lang.bind(this, this._onLeave));
         this._havePointer = false;
 
-        this._draggable = DND.makeDraggable(this.actor);
+        this._draggable = DND.makeDraggable(this.actor,
+                                            { dragActorMaxSize: WINDOW_DND_SIZE,
+                                              dragActorOpacity: DRAGGING_WINDOW_OPACITY });
         this._draggable.connect('drag-begin', Lang.bind(this, this._onDragBegin));
         this._draggable.connect('drag-end', Lang.bind(this, this._onDragEnd));
-        this._inDrag = false;
+        this.inDrag = false;
 
         this._zooming = false;
         this._selected = false;
@@ -135,7 +141,7 @@ WindowClone.prototype = {
 
     setStackAbove: function (actor) {
         this._stackAbove = actor;
-        if (this._inDrag || this._zooming)
+        if (this.inDrag || this._zooming)
             // We'll fix up the stack after the drag/zooming
             return;
         this.actor.raise(this._stackAbove);
@@ -160,14 +166,23 @@ WindowClone.prototype = {
     },
 
     _onDestroy: function() {
+        this.metaWindow._delegate = null;
+        this.actor._delegate = null;
         if (this._zoomLightbox)
             this._zoomLightbox.destroy();
+
+        if (this.inDrag) {
+            this.emit('drag-end');
+            this.inDrag = false;
+        }
+
+        this.disconnectAll();
     },
 
     _onEnter: function (actor, event) {
         // If the user drags faster than we can follow, he'll end up
         // leaving the window temporarily and then re-entering it
-        if (this._inDrag)
+        if (this.inDrag)
             return;
 
         this._havePointer = true;
@@ -176,7 +191,7 @@ WindowClone.prototype = {
     _onLeave: function (actor, event) {
         // If the user drags faster than we can follow, he'll end up
         // leaving the window temporarily and then re-entering it
-        if (this._inDrag)
+        if (this.inDrag)
             return;
 
         this._havePointer = false;
@@ -276,12 +291,12 @@ WindowClone.prototype = {
     },
 
     _onDragBegin : function (draggable, time) {
-        this._inDrag = true;
+        this.inDrag = true;
         this.emit('drag-begin');
     },
 
     _onDragEnd : function (draggable, time, snapback) {
-        this._inDrag = false;
+        this.inDrag = false;
 
         // Most likely, the clone is going to move away from the
         // pointer now. But that won't cause a leave-event, so
@@ -589,32 +604,30 @@ const WindowPositionFlags = {
 };
 
 /**
- * @workspaceNum: Workspace index
- * @parentActor: The actor which will be the parent of this workspace;
- *               we need this in order to add chrome such as the icons
- *               on top of the windows without having them be scaled.
+ * @metaWorkspace: a #Meta.Workspace
  */
-function Workspace(workspaceNum, parentActor) {
-    this._init(workspaceNum, parentActor);
+function Workspace(metaWorkspace) {
+    this._init(metaWorkspace);
 }
 
 Workspace.prototype = {
-    _init : function(workspaceNum, parentActor) {
-        this.workspaceNum = workspaceNum;
+    _init : function(metaWorkspace) {
+        // When dragging a window, we use this slot for reserve space.
+        this._reservedSlot = null;
+        this.metaWorkspace = metaWorkspace;
         this._windowOverlaysGroup = new Clutter.Group();
         // Without this the drop area will be overlapped.
         this._windowOverlaysGroup.set_size(0, 0);
 
-        this._metaWorkspace = global.screen.get_workspace_by_index(workspaceNum);
-
-        parentActor.add_actor(this._windowOverlaysGroup);
-        this._parentActor = parentActor;
-
         this.actor = new Clutter.Group();
         this.actor._delegate = this;
 
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
+        // Items in _windowOverlaysGroup should not be scaled, so we don't
+        // add them to this.actor, but to its parent whenever it changes
+        this.actor.connect('parent-set', Lang.bind(this, this._onParentSet));
+
         // Auto-sizing is unreliable in the presence of ClutterClone, so rather than
         // implicitly counting on the workspace actor to be sized to the size of the
         // included desktop actor clone, set the size explicitly to the screen size.
@@ -639,8 +652,13 @@ Workspace.prototype = {
         this._desktop.connect('selected',
                               Lang.bind(this,
                                         function(clone, time) {
-                                            this._metaWorkspace.activate(time);
-                                            Main.overview.hide();
+                                            // Only switch to the workspace when one window is open
+                                            // We check for "2" here because the Desktop does not really count
+                                            // as a window in this context.
+                                            if (this._windows.length <= 2) {
+                                                this.metaWorkspace.activate(time);
+                                                Main.overview.hide();
+                                            }
                                         }));
         this.actor.add_actor(this._desktop.actor);
 
@@ -658,10 +676,11 @@ Workspace.prototype = {
         this._showOnlyWindows = null;
 
         // Track window changes
-        this._windowAddedId = this._metaWorkspace.connect('window-added',
+        this._windowAddedId = this.metaWorkspace.connect('window-added',
                                                           Lang.bind(this, this._windowAdded));
-        this._windowRemovedId = this._metaWorkspace.connect('window-removed',
+        this._windowRemovedId = this.metaWorkspace.connect('window-removed',
                                                             Lang.bind(this, this._windowRemoved));
+        this._repositionWindowsId = 0;
 
         this._visible = false;
 
@@ -679,6 +698,17 @@ Workspace.prototype = {
         return -1;
     },
 
+    _onParentSet: function(actor, old_parent) {
+        let new_parent = this.actor.get_parent();
+        if (new_parent == null)
+            return;
+
+        if (old_parent)
+            this._windowOverlaysGroup.reparent(new_parent);
+        else
+            new_parent.add_actor(this._windowOverlaysGroup);
+    },
+
     /**
      * lookupCloneForMetaWindow:
      * @metaWindow: A #MetaWindow
@@ -761,6 +791,16 @@ Workspace.prototype = {
         }
     },
 
+    /**
+     * setReactive:
+     * @reactive: %true iff the workspace should be reactive
+     *
+     * Set the workspace (desktop) reactive
+     **/
+    setReactive: function(reactive) {
+        this._desktop.actor.reactive = reactive;
+    },
+
     _updateFramePosition : function() {
         this._frame.set_position(this._desktop.actor.x - FRAME_SIZE / this.actor.scale_x,
                                  this._desktop.actor.y - FRAME_SIZE / this.actor.scale_y);
@@ -855,7 +895,7 @@ Workspace.prototype = {
 
     /**
      * _computeWindowMotion:
-     * @metaWindow: A #MetaWindow
+     * @actor: A #WindowClone's #ClutterActor
      * @slot: An element of #POSITIONS
      * @slotGeometry: Layout of @slot
      *
@@ -864,15 +904,15 @@ Workspace.prototype = {
      * Currently this is the square of the distance between the
      * centers.
      */
-    _computeWindowMotion: function (metaWindow, slot) {
-        let rect = new Meta.Rectangle();
-        metaWindow.get_outer_rect(rect);
-
+    _computeWindowMotion: function (actor, slot) {
         let [xCenter, yCenter, fraction] = slot;
         let xDelta, yDelta, distanceSquared;
+        let actorWidth, actorHeight;
 
-        xDelta = rect.x + rect.width / 2.0 - xCenter * global.screen_width;
-        yDelta = rect.y + rect.height / 2.0 - yCenter * global.screen_height;
+        actorWidth = actor.width * actor.scale_x;
+        actorHeight = actor.height * actor.scale_y;
+        xDelta = actor.x + actorWidth / 2.0 - xCenter * global.screen_width;
+        yDelta = actor.y + actorHeight / 2.0 - yCenter * global.screen_height;
         distanceSquared = xDelta * xDelta + yDelta * yDelta;
 
         return distanceSquared;
@@ -891,10 +931,10 @@ Workspace.prototype = {
         this._forEachPermutations(clones, Lang.bind(this, function (permutation) {
             let motion = 0;
             for (let i = 0; i < permutation.length; i++) {
-                let metaWindow = permutation[i].metaWindow;
+                let cloneActor = permutation[i].actor;
                 let slot = slots[i];
 
-                let delta = this._computeWindowMotion(metaWindow, slot);
+                let delta = this._computeWindowMotion(cloneActor, slot);
 
                 motion += delta;
             }
@@ -924,8 +964,8 @@ Workspace.prototype = {
             let minimumMotionIndex = -1;
             let minimumMotion = -1;
             for (let j = 0; j < clonesCopy.length; j++) {
-                let metaWindow = clonesCopy[j].metaWindow;
-                let delta = this._computeWindowMotion(metaWindow, slot);
+                let cloneActor = clonesCopy[j].actor;
+                let delta = this._computeWindowMotion(cloneActor, slot);
                 if (minimumMotionIndex == -1 || delta < minimumMotion) {
                     minimumMotionIndex = j;
                     minimumMotion = delta;
@@ -992,11 +1032,16 @@ Workspace.prototype = {
 
         let rect = new Meta.Rectangle();
         metaWindow.get_outer_rect(rect);
+        let buttonOuterHeight, captionHeight;
+        let buttonOuterWidth = 0;
 
-        let [buttonOuterHeight, captionHeight] = this._windowOverlays[1].chromeHeights();
+        if (this._windowOverlays[1]) {
+            [buttonOuterHeight, captionHeight] = this._windowOverlays[1].chromeHeights();
+            buttonOuterWidth = this._windowOverlays[1].chromeWidth() / this.scale;
+        } else
+            [buttonOuterHeight, captionHeight] = [0, 0];
         buttonOuterHeight /= this.scale;
         captionHeight /= this.scale;
-        let buttonOuterWidth = this._windowOverlays[1].chromeWidth() / this.scale;
 
         let desiredWidth = global.screen_width * fraction;
         let desiredHeight = global.screen_height * fraction;
@@ -1010,6 +1055,19 @@ Workspace.prototype = {
         return [x, y, scale];
     },
 
+    setReservedSlot: function(clone) {
+        if (clone && this.containsMetaWindow(clone.metaWindow)) {
+            this._reservedSlot = null;
+            this.positionWindows(WindowPositionFlags.ANIMATE);
+            return;
+        }
+        if (clone)
+            this._reservedSlot = clone;
+        else
+            this._reservedSlot = null;
+        this.positionWindows(WindowPositionFlags.ANIMATE);
+    },
+
     /**
      * positionWindows:
      * @flags:
@@ -1017,9 +1075,16 @@ Workspace.prototype = {
      *  ANIMATE - Indicates that we need animate changing position.
      */
     positionWindows : function(flags) {
+        if (this._repositionWindowsId > 0) {
+            Mainloop.source_remove(this._repositionWindowsId);
+            this._repositionWindowsId = 0;
+        }
+
         let totalVisible = 0;
 
         let visibleClones = this._getVisibleClones();
+        if (this._reservedSlot)
+            visibleClones.push(this._reservedSlot);
 
         let workspaceZooming = flags & WindowPositionFlags.ZOOM;
         let animate = flags & WindowPositionFlags.ANIMATE;
@@ -1035,9 +1100,15 @@ Workspace.prototype = {
             let mainIndex = this._lookupIndex(metaWindow);
             let overlay = this._windowOverlays[mainIndex];
 
+            // Positioning a window currently being dragged must be avoided;
+            // we'll just leave a blank spot in the layout for it.
+            if (clone.inDrag)
+                continue;
+
             let [x, y, scale] = this._computeWindowRelativeLayout(metaWindow, slot);
 
-            overlay.hide();
+            if (overlay)
+                overlay.hide();
             if (animate) {
                 if (!metaWindow.showing_on_its_workspace()) {
                     /* Hidden windows should fade in and grow
@@ -1111,8 +1182,13 @@ Workspace.prototype = {
         cloneWidth = this.scale * clone.actor.scale_x * cloneWidth;
         cloneHeight = this.scale * clone.actor.scale_y * cloneHeight;
 
-        overlay.updatePositions(cloneX, cloneY, cloneWidth, cloneHeight);
-        overlay.fadeIn();
+        if (!this._windowOverlaysGroup.visible)
+            this._windowOverlaysGroup.show();
+
+        if (overlay) {
+            overlay.updatePositions(cloneX, cloneY, cloneWidth, cloneHeight);
+            overlay.fadeIn();
+        }
     },
 
     _fadeInAllOverlays: function() {
@@ -1132,6 +1208,36 @@ Workspace.prototype = {
         }
     },
 
+    _delayedWindowRepositioning: function() {
+        let [child, x, y, mask] =
+            Gdk.Screen.get_default().get_root_window().get_pointer();
+        let wsWidth = this.actor.width * this.scale;
+        let wsHeight = this.actor.height * this.scale;
+
+        let pointerHasMoved = (this._cursorX != x && this._cursorY != y);
+        let inWorkspace = (this.gridX < x && x < this.gridX + wsWidth &&
+                           this.gridY < y && y < this.gridY + wsHeight);
+
+        if (pointerHasMoved && inWorkspace) {
+            // store current cursor position
+            this._cursorX = x;
+            this._cursorY = y;
+            return true;
+        }
+
+        this.positionWindows(WindowPositionFlags.ANIMATE);
+        return false;
+    },
+
+    showWindowsOverlays: function() {
+        this._windowOverlaysGroup.show();
+        this._fadeInAllOverlays();
+    },
+
+    hideWindowsOverlays: function() {
+        this._windowOverlaysGroup.hide();
+    },
+
     _windowRemoved : function(metaWorkspace, metaWin) {
         let win = metaWin.get_compositor_private();
 
@@ -1163,7 +1269,26 @@ Workspace.prototype = {
         }
         clone.destroy();
 
-        this.positionWindows(WindowPositionFlags.ANIMATE);
+
+        // We need to reposition the windows; to avoid shuffling windows
+        // around while the user is interacting with the workspace, we delay
+        // the positioning until the pointer remains still for at least 750 ms
+        // or is moved outside the workspace
+
+        // remove old handler
+        if (this._repositionWindowsId > 0) {
+            Mainloop.source_remove(this._repositionWindowsId);
+            this._repositionWindowsId = 0;
+        }
+
+        // setup new handler
+        let [child, x, y, mask] =
+            Gdk.Screen.get_default().get_root_window().get_pointer();
+        this._cursorX = x;
+        this._cursorY = y;
+
+        this._repositionWindowsId = Mainloop.timeout_add(750,
+            Lang.bind(this, this._delayedWindowRepositioning));
     },
 
     _windowAdded : function(metaWorkspace, metaWin) {
@@ -1215,19 +1340,19 @@ Workspace.prototype = {
     },
 
     // Animate the full-screen to Overview transition.
-    zoomToOverview : function(animate) {
+    zoomToOverview : function() {
         this.actor.set_position(this.gridX, this.gridY);
         this.actor.set_scale(this.scale, this.scale);
 
         // Position and scale the windows.
-        if (animate)
+        if (Main.overview.animationInProgress)
             this.positionWindows(WindowPositionFlags.ANIMATE | WindowPositionFlags.ZOOM);
         else
             this.positionWindows(WindowPositionFlags.ZOOM);
 
-        let active = global.screen.get_active_workspace_index();
-        let fadeInIcons = (animate &&
-                           active == this.workspaceNum &&
+        let active = global.screen.get_active_workspace();
+        let fadeInIcons = (Main.overview.animationInProgress &&
+                           active == this.metaWorkspace &&
                            !this._haveMaximizedWindows());
         this._desktop.zoomToOverview(fadeInIcons);
 
@@ -1240,8 +1365,12 @@ Workspace.prototype = {
 
         this._hideAllOverlays();
 
-        Main.overview.connect('hidden', Lang.bind(this,
-                                                  this._doneLeavingOverview));
+        if (this._repositionWindowsId > 0) {
+            Mainloop.source_remove(this._repositionWindowsId);
+            this._repositionWindowsId = 0;
+        }
+        this._overviewHiddenId = Main.overview.connect('hidden', Lang.bind(this,
+                                                                           this._doneLeavingOverview));
 
         // Position and scale the windows.
         for (let i = 1; i < this._windows.length; i++) {
@@ -1273,8 +1402,8 @@ Workspace.prototype = {
             }
         }
 
-        let active = global.screen.get_active_workspace_index();
-        let fadeOutIcons = (active == this.workspaceNum &&
+        let active = global.screen.get_active_workspace();
+        let fadeOutIcons = (active == this.metaWorkspace &&
                             !this._haveMaximizedWindows());
         this._desktop.zoomFromOverview(fadeOutIcons);
 
@@ -1349,10 +1478,17 @@ Workspace.prototype = {
     },
 
     _onDestroy: function(actor) {
+        if (this._overviewHiddenId) {
+            Main.overview.disconnect(this._overviewHiddenId);
+            this._overviewHiddenId = 0;
+        }
         Tweener.removeTweens(actor);
 
-        this._metaWorkspace.disconnect(this._windowAddedId);
-        this._metaWorkspace.disconnect(this._windowRemovedId);
+        this.metaWorkspace.disconnect(this._windowAddedId);
+        this.metaWorkspace.disconnect(this._windowRemovedId);
+
+        if (this._repositionWindowsId > 0)
+            Mainloop.source_remove(this._repositionWindowsId);
 
         // Usually, the windows will be destroyed automatically with
         // their parent (this.actor), but we might have a zoomed window
@@ -1360,6 +1496,7 @@ Workspace.prototype = {
         // the desktop window, which is never reparented
         for (let w = 1; w < this._windows.length; w++)
             this._windows[w].destroy();
+        this._windows = [];
     },
 
     // Sets this.leavingOverview flag to false.
@@ -1369,13 +1506,13 @@ Workspace.prototype = {
 
     // Tests if @win belongs to this workspaces
     _isMyWindow : function (win) {
-        return win.get_workspace() == this.workspaceNum ||
+        return win.get_workspace() == this.metaWorkspace.index() ||
             (win.get_meta_window() && win.get_meta_window().is_on_all_workspaces());
     },
 
     // Tests if @win should be shown in the Overview
     _isOverviewWindow : function (win) {
-        let tracker = Shell.WindowTracker.get_default()
+        let tracker = Shell.WindowTracker.get_default();
         return tracker.is_window_interesting(win.get_meta_window());
     },
 
@@ -1387,11 +1524,13 @@ Workspace.prototype = {
         clone.connect('selected',
                       Lang.bind(this, this._onCloneSelected));
         clone.connect('drag-begin',
-                      Lang.bind(this, function() {
+                      Lang.bind(this, function(clone) {
+                          this.emit('window-drag-begin', clone.actor);
                           overlay.hide();
                       }));
         clone.connect('drag-end',
-                      Lang.bind(this, function() {
+                      Lang.bind(this, function(clone) {
+                          this.emit('window-drag-end', clone.actor);
                           overlay.show();
                       }));
 
@@ -1440,14 +1579,12 @@ Workspace.prototype = {
     },
 
     _onCloneSelected : function (clone, time) {
-        Main.activateWindow(clone.metaWindow, time, this.workspaceNum);
+        Main.activateWindow(clone.metaWindow, time,
+                            this.metaWorkspace.index());
     },
 
     _removeSelf : function(actor, event) {
-        let screen = global.screen;
-        let workspace = screen.get_workspace_by_index(this.workspaceNum);
-
-        screen.remove_workspace(workspace, event.get_time());
+        screen.remove_workspace(this.metaWorkspace, event.get_time());
         return true;
     },
 
@@ -1467,12 +1604,12 @@ Workspace.prototype = {
             };
 
             let metaWindow = win.get_meta_window();
-            metaWindow.change_workspace_by_index(this.workspaceNum,
+            metaWindow.change_workspace_by_index(this.metaWorkspace.index(),
                                                  false, // don't create workspace
                                                  time);
             return true;
         } else if (source.shellWorkspaceLaunch) {
-            this._metaWorkspace.activate(time);
+            this.metaWorkspace.activate(time);
             source.shellWorkspaceLaunch();
             return true;
         }
