@@ -1,5 +1,7 @@
 /* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
 
+#include "config.h"
+
 #include "shell-app-system.h"
 #include <string.h>
 
@@ -10,8 +12,9 @@
 #include <gconf/gconf-client.h>
 #include <clutter/clutter.h>
 
+#include "shell-app-private.h"
 #include "shell-global.h"
-#include "shell-texture-cache.h"
+#include "st/st-texture-cache.h"
 #include "display.h"
 
 #define GMENU_I_KNOW_THIS_IS_UNSTABLE
@@ -33,7 +36,6 @@ enum {
 
 enum {
   INSTALLED_CHANGED,
-  FAVORITES_CHANGED,
   LAST_SIGNAL
 };
 
@@ -43,25 +45,21 @@ struct _ShellAppSystemPrivate {
   GMenuTree *apps_tree;
   GMenuTree *settings_tree;
 
+  GHashTable *app_id_to_info;
   GHashTable *app_id_to_app;
 
-  GSList *cached_app_menus; /* ShellAppMenuEntry */
-
+  GSList *cached_flattened_apps; /* ShellAppInfo */
   GSList *cached_settings; /* ShellAppInfo */
 
-  GList *cached_favorites; /* utf8 */
-
   gint app_monitor_id;
 
   guint app_change_timeout_id;
 };
 
 static void shell_app_system_finalize (GObject *object);
-static void on_tree_changed (gpointer user_data);
+static gboolean on_tree_changed (gpointer user_data);
 static void on_tree_changed_cb (GMenuTree *tree, gpointer user_data);
 static void reread_menus (ShellAppSystem *self);
-static void on_favorite_apps_changed (GConfClient *client, guint id, GConfEntry *entry, gpointer user_data);
-static void reread_favorite_apps (ShellAppSystem *system);
 
 G_DEFINE_TYPE(ShellAppSystem, shell_app_system, G_TYPE_OBJECT);
 
@@ -82,6 +80,11 @@ struct _ShellAppInfo {
    */
   guint refcount;
 
+  char *casefolded_name;
+  char *name_collation_key;
+  char *casefolded_description;
+  char *casefolded_exec;
+
   GMenuTreeItem *entry;
 
   GKeyFile *keyfile;
@@ -103,6 +106,11 @@ shell_app_info_unref (ShellAppInfo *info)
 {
   if (--info->refcount > 0)
     return;
+
+  g_free (info->casefolded_name);
+  g_free (info->name_collation_key);
+  g_free (info->casefolded_description);
+
   switch (info->type)
   {
   case SHELL_APP_INFO_TYPE_ENTRY:
@@ -128,7 +136,7 @@ shell_app_info_new_from_tree_item (GMenuTreeItem *item)
   if (!item)
     return NULL;
 
-  info = g_slice_alloc (sizeof (ShellAppInfo));
+  info = g_slice_alloc0 (sizeof (ShellAppInfo));
   info->type = SHELL_APP_INFO_TYPE_ENTRY;
   info->refcount = 1;
   info->entry = gmenu_tree_item_ref (item);
@@ -140,7 +148,7 @@ shell_app_info_new_from_window (MetaWindow *window)
 {
   ShellAppInfo *info;
 
-  info = g_slice_alloc (sizeof (ShellAppInfo));
+  info = g_slice_alloc0 (sizeof (ShellAppInfo));
   info->type = SHELL_APP_INFO_TYPE_WINDOW;
   info->refcount = 1;
   info->window = g_object_ref (window);
@@ -158,7 +166,7 @@ shell_app_info_new_from_keyfile_take_ownership (GKeyFile   *keyfile,
 {
   ShellAppInfo *info;
 
-  info = g_slice_alloc (sizeof (ShellAppInfo));
+  info = g_slice_alloc0 (sizeof (ShellAppInfo));
   info->type = SHELL_APP_INFO_TYPE_DESKTOP_FILE;
   info->refcount = 1;
   info->keyfile = keyfile;
@@ -166,29 +174,6 @@ shell_app_info_new_from_keyfile_take_ownership (GKeyFile   *keyfile,
   return info;
 }
 
-static gpointer
-shell_app_menu_entry_copy (gpointer entryp)
-{
-  ShellAppMenuEntry *entry;
-  ShellAppMenuEntry *copy;
-  entry = entryp;
-  copy = g_new0 (ShellAppMenuEntry, 1);
-  copy->name = g_strdup (entry->name);
-  copy->id = g_strdup (entry->id);
-  copy->icon = g_strdup (entry->icon);
-  return copy;
-}
-
-static void
-shell_app_menu_entry_free (gpointer entryp)
-{
-  ShellAppMenuEntry *entry = entryp;
-  g_free (entry->name);
-  g_free (entry->id);
-  g_free (entry->icon);
-  g_free (entry);
-}
-
 static void shell_app_system_class_init(ShellAppSystemClass *klass)
 {
   GObjectClass *gobject_class = (GObjectClass *)klass;
@@ -203,14 +188,6 @@ static void shell_app_system_class_init(ShellAppSystemClass *klass)
 		  NULL, NULL,
 		  g_cclosure_marshal_VOID__VOID,
 		  G_TYPE_NONE, 0);
-  signals[FAVORITES_CHANGED] =
-    g_signal_new ("favorites-changed",
-                  SHELL_TYPE_APP_SYSTEM,
-                  G_SIGNAL_RUN_LAST,
-                  G_STRUCT_OFFSET (ShellAppSystemClass, favorites_changed),
-                  NULL, NULL,
-                  g_cclosure_marshal_VOID__VOID,
-                  G_TYPE_NONE, 0);
 
   g_type_class_add_private (gobject_class, sizeof (ShellAppSystemPrivate));
 }
@@ -226,8 +203,11 @@ shell_app_system_init (ShellAppSystem *self)
                                                    ShellAppSystemPrivate);
 
   /* The key is owned by the value */
-  priv->app_id_to_app = g_hash_table_new_full (g_str_hash, g_str_equal,
-                                               NULL, (GDestroyNotify) shell_app_info_unref);
+  priv->app_id_to_info = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                                NULL, (GDestroyNotify) shell_app_info_unref);
+
+  /* Key is owned by info */
+  priv->app_id_to_app = g_hash_table_new (g_str_hash, g_str_equal);
 
   /* For now, we want to pick up Evince, Nautilus, etc.  We'll
    * handle NODISPLAY semantics at a higher level or investigate them
@@ -244,10 +224,6 @@ shell_app_system_init (ShellAppSystem *self)
   reread_menus (self);
 
   client = gconf_client_get_default ();
-
-  self->priv->app_monitor_id = gconf_client_notify_add (client, SHELL_APP_FAVORITES_KEY,
-                                                        on_favorite_apps_changed, self, NULL, NULL);
-  reread_favorite_apps (self);
 }
 
 static void
@@ -262,69 +238,23 @@ shell_app_system_finalize (GObject *object)
   gmenu_tree_unref (priv->apps_tree);
   gmenu_tree_unref (priv->settings_tree);
 
+  g_hash_table_destroy (priv->app_id_to_info);
   g_hash_table_destroy (priv->app_id_to_app);
 
-  g_slist_foreach (priv->cached_app_menus, (GFunc)shell_app_menu_entry_free, NULL);
-  g_slist_free (priv->cached_app_menus);
-  priv->cached_app_menus = NULL;
-
+  g_slist_foreach (priv->cached_flattened_apps, (GFunc)shell_app_info_unref, NULL);
+  g_slist_free (priv->cached_flattened_apps);
+  priv->cached_flattened_apps = NULL;
   g_slist_foreach (priv->cached_settings, (GFunc)shell_app_info_unref, NULL);
   g_slist_free (priv->cached_settings);
   priv->cached_settings = NULL;
 
-  g_list_free (priv->cached_favorites);
-
-  gconf_client_notify_remove (gconf_client_get_default (), priv->app_monitor_id);
-
   G_OBJECT_CLASS (shell_app_system_parent_class)->finalize(object);
 }
 
-static void
-reread_directories (ShellAppSystem *self, GSList **cache, GMenuTree *tree)
-{
-  GMenuTreeDirectory *trunk;
-  GSList *entries;
-  GSList *iter;
-
-  trunk = gmenu_tree_get_root_directory (tree);
-  entries = gmenu_tree_directory_get_contents (trunk);
-
-  g_slist_foreach (*cache, (GFunc)shell_app_menu_entry_free, NULL);
-  g_slist_free (*cache);
-  *cache = NULL;
-
-  for (iter = entries; iter; iter = iter->next)
-    {
-      GMenuTreeItem *item = iter->data;
-
-      switch (gmenu_tree_item_get_type (item))
-        {
-          case GMENU_TREE_ITEM_DIRECTORY:
-            {
-              GMenuTreeDirectory *dir = iter->data;
-              ShellAppMenuEntry *shell_entry = g_new0 (ShellAppMenuEntry, 1);
-              shell_entry->name = g_strdup (gmenu_tree_directory_get_name (dir));
-              shell_entry->id = g_strdup (gmenu_tree_directory_get_menu_id (dir));
-              shell_entry->icon = g_strdup (gmenu_tree_directory_get_icon (dir));
-
-              *cache = g_slist_prepend (*cache, shell_entry);
-            }
-            break;
-          default:
-            break;
-        }
-
-      gmenu_tree_item_unref (item);
-    }
-  *cache = g_slist_reverse (*cache);
-
-  g_slist_free (entries);
-  gmenu_tree_item_unref (trunk);
-}
-
 static GSList *
 gather_entries_recurse (ShellAppSystem     *monitor,
                         GSList             *apps,
+                        GHashTable         *unique,
                         GMenuTreeDirectory *root)
 {
   GSList *contents;
@@ -340,13 +270,17 @@ gather_entries_recurse (ShellAppSystem     *monitor,
           case GMENU_TREE_ITEM_ENTRY:
             {
               ShellAppInfo *app = shell_app_info_new_from_tree_item (item);
-              apps = g_slist_prepend (apps, app);
+              if (!g_hash_table_lookup (unique, shell_app_info_get_id (app)))
+                {
+                  apps = g_slist_prepend (apps, app);
+                  g_hash_table_insert (unique, (char*)shell_app_info_get_id (app), app);
+                }
             }
             break;
           case GMENU_TREE_ITEM_DIRECTORY:
             {
               GMenuTreeDirectory *dir = (GMenuTreeDirectory*)item;
-              apps = gather_entries_recurse (monitor, apps, dir);
+              apps = gather_entries_recurse (monitor, apps, unique, dir);
             }
             break;
           default:
@@ -363,6 +297,7 @@ gather_entries_recurse (ShellAppSystem     *monitor,
 static void
 reread_entries (ShellAppSystem     *self,
                 GSList            **cache,
+                GHashTable         *unique,
                 GMenuTree          *tree)
 {
   GMenuTreeDirectory *trunk;
@@ -373,55 +308,53 @@ reread_entries (ShellAppSystem     *self,
   g_slist_free (*cache);
   *cache = NULL;
 
-  *cache = gather_entries_recurse (self, *cache, trunk);
+  *cache = gather_entries_recurse (self, *cache, unique, trunk);
 
   gmenu_tree_item_unref (trunk);
 }
 
 static void
-cache_by_id (ShellAppSystem *self, GSList *apps, gboolean ref)
+cache_by_id (ShellAppSystem *self, GSList *apps)
 {
   GSList *iter;
 
   for (iter = apps; iter; iter = iter->next)
     {
       ShellAppInfo *info = iter->data;
-      if (ref)
-        shell_app_info_ref (info);
+      shell_app_info_ref (info);
       /* the name is owned by the info itself */
-      g_hash_table_insert (self->priv->app_id_to_app, (char*)shell_app_info_get_id (info),
-                           info);
+      g_hash_table_replace (self->priv->app_id_to_info, (char*)shell_app_info_get_id (info),
+                            info);
     }
 }
 
 static void
 reread_menus (ShellAppSystem *self)
 {
-  GSList *apps;
-  GMenuTreeDirectory *trunk;
+  GHashTable *unique = g_hash_table_new (g_str_hash, g_str_equal);
 
-  reread_directories (self, &(self->priv->cached_app_menus), self->priv->apps_tree);
+  reread_entries (self, &(self->priv->cached_flattened_apps), unique, self->priv->apps_tree);
+  g_hash_table_remove_all (unique);
+  reread_entries (self, &(self->priv->cached_settings), unique, self->priv->settings_tree);
+  g_hash_table_destroy (unique);
 
-  reread_entries (self, &(self->priv->cached_settings), self->priv->settings_tree);
+  g_hash_table_remove_all (self->priv->app_id_to_info);
 
-  /* Now loop over applications.menu and settings.menu, inserting each by desktop file
-   * ID into a hash */
-  g_hash_table_remove_all (self->priv->app_id_to_app);
-  trunk = gmenu_tree_get_root_directory (self->priv->apps_tree);
-  apps = gather_entries_recurse (self, NULL, trunk);
-  gmenu_tree_item_unref (trunk);
-  cache_by_id (self, apps, FALSE);
-  g_slist_free (apps);
-  cache_by_id (self, self->priv->cached_settings, TRUE);
+  cache_by_id (self, self->priv->cached_flattened_apps);
+  cache_by_id (self, self->priv->cached_settings);
 }
 
-static void
+static gboolean
 on_tree_changed (gpointer user_data)
 {
   ShellAppSystem *self = SHELL_APP_SYSTEM (user_data);
-  g_signal_emit (self, signals[INSTALLED_CHANGED], 0);
+
   reread_menus (self);
+
+  g_signal_emit (self, signals[INSTALLED_CHANGED], 0);
+
   self->priv->app_change_timeout_id = 0;
+  return FALSE;
 }
 
 static void
@@ -449,63 +382,6 @@ on_tree_changed_cb (GMenuTree *monitor, gpointer user_data)
                                                           self, NULL);
 }
 
-static GList *
-convert_gconf_value_string_list_to_list_uniquify (GConfValue *value )
-{
-  GSList *list;
-  GSList *tmp;
-  GList *result = NULL;
-  GHashTable *tmp_table = g_hash_table_new (g_str_hash, g_str_equal);
-
-  list = gconf_value_get_list (value);
-
-  for (tmp = list ; tmp; tmp = tmp->next)
-    {
-      GConfValue *value = tmp->data;
-      char *str = g_strdup (gconf_value_get_string (value));
-      if (!str)
-        continue;
-      if (g_hash_table_lookup (tmp_table, str))
-        {
-          g_free (str);
-          continue;
-        }
-      g_hash_table_insert (tmp_table, str, GUINT_TO_POINTER(1));
-      result = g_list_prepend (result, str);
-    }
-  g_hash_table_destroy (tmp_table);
-  return g_list_reverse (result);
-}
-
-static void
-reread_favorite_apps (ShellAppSystem *system)
-{
-  GConfClient *client = gconf_client_get_default ();
-  GConfValue *val;
-
-  val = gconf_client_get (client, SHELL_APP_FAVORITES_KEY, NULL);
-
-  if (!(val && val->type == GCONF_VALUE_LIST && gconf_value_get_list_type (val) == GCONF_VALUE_STRING))
-    return;
-
-  g_list_foreach (system->priv->cached_favorites, (GFunc) g_free, NULL);
-  g_list_free (system->priv->cached_favorites);
-  system->priv->cached_favorites = convert_gconf_value_string_list_to_list_uniquify (val);
-
-  gconf_value_free (val);
-}
-
-void
-on_favorite_apps_changed (GConfClient *client,
-                          guint        id,
-                          GConfEntry  *entry,
-                          gpointer     user_data)
-{
-  ShellAppSystem *system = SHELL_APP_SYSTEM (user_data);
-  reread_favorite_apps (system);
-  g_signal_emit (G_OBJECT (system), signals[FAVORITES_CHANGED], 0);
-}
-
 GType
 shell_app_info_get_type (void)
 {
@@ -519,58 +395,19 @@ shell_app_info_get_type (void)
   return gtype;
 }
 
-GType
-shell_app_menu_entry_get_type (void)
-{
-  static GType gtype = G_TYPE_INVALID;
-  if (gtype == G_TYPE_INVALID)
-    {
-      gtype = g_boxed_type_register_static ("ShellAppMenuEntry",
-          shell_app_menu_entry_copy,
-          shell_app_menu_entry_free);
-    }
-  return gtype;
-}
-
 /**
- * shell_app_system_get_applications_for_menu:
+ * shell_app_system_get_flattened_apps:
  *
  * Traverses a toplevel menu, and returns all items under it.  Nested items
- * are flattened.
+ * are flattened.  This value is computed on initial call and cached thereafter
+ * until the set of installed applications changes.
  *
- * Return value: (transfer container) (element-type ShellAppInfo): List of applications
+ * Return value: (transfer none) (element-type ShellAppInfo): List of applications
  */
 GSList *
-shell_app_system_get_applications_for_menu (ShellAppSystem *monitor,
-                                            const char *menu)
+shell_app_system_get_flattened_apps (ShellAppSystem *self)
 {
-  char *path;
-  GMenuTreeDirectory *menu_entry;
-  GSList *apps;
-
-  path = g_strdup_printf ("/%s", menu);
-  menu_entry = gmenu_tree_get_directory_from_path (monitor->priv->apps_tree, path);
-  g_free (path);
-  g_assert (menu_entry != NULL);
-
-  apps = gather_entries_recurse (monitor, NULL, menu_entry);
-
-  gmenu_tree_item_unref (menu_entry);
-
-  return apps;
-}
-
-/**
- * shell_app_system_get_menus:
- *
- * Returns a list of toplevel #ShellAppMenuEntry items
- *
- * Return value: (transfer none) (element-type AppMenuEntry): List of toplevel menus
- */
-GSList *
-shell_app_system_get_menus (ShellAppSystem *monitor)
-{
-  return monitor->priv->cached_app_menus;
+  return self->priv->cached_flattened_apps;
 }
 
 /**
@@ -602,93 +439,96 @@ shell_app_system_get_default ()
   return instance;
 }
 
-/**
- * shell_app_system_get_favorites:
- *
- * Return the list of applications which have been explicitly added to the
- * favorites.
- *
- * Return value: (transfer none) (element-type utf8): List of favorite application ids
- */
-GList *
-shell_app_system_get_favorites (ShellAppSystem *system)
-{
-  return system->priv->cached_favorites;
-}
+typedef struct {
+  ShellAppSystem *appsys;
+  ShellAppInfo *info;
+} ShellAppRef;
 
 static void
-set_gconf_value_string_list (GConfValue *val, GList *items)
+shell_app_system_on_app_weakref (gpointer  data,
+                                 GObject  *location)
 {
-  GList *iter;
-  GSList *tmp = NULL;
+  ShellAppRef *ref = data;
 
-  for (iter = items; iter; iter = iter->next)
-    {
-      const char *str = iter->data;
-      GConfValue *strval = gconf_value_new (GCONF_VALUE_STRING);
-      gconf_value_set_string (strval, str);
-      tmp = g_slist_prepend (tmp, strval);
-    }
-  tmp = g_slist_reverse (tmp);
-
-  gconf_value_set_list (val, tmp);
-  g_slist_free (tmp);
-}
-
-void
-shell_app_system_add_favorite (ShellAppSystem *system, const char *id)
-{
-  GConfClient *client = gconf_client_get_default ();
-  GConfValue *val;
-  GList *iter;
-
-  iter = g_list_find_custom (system->priv->cached_favorites, id, (GCompareFunc)strcmp);
-  if (iter)
-    return;
-
-  val = gconf_value_new (GCONF_VALUE_LIST);
-  gconf_value_set_list_type (val, GCONF_VALUE_STRING);
-
-  system->priv->cached_favorites = g_list_append (system->priv->cached_favorites, g_strdup (id));
-
-  set_gconf_value_string_list (val, system->priv->cached_favorites);
-  gconf_client_set (client, SHELL_APP_FAVORITES_KEY, val, NULL);
-}
-
-void
-shell_app_system_remove_favorite (ShellAppSystem *system, const char *id)
-{
-  GConfClient *client = gconf_client_get_default ();
-  GConfValue *val;
-  GList *iter;
-
-  iter = g_list_find_custom (system->priv->cached_favorites, id, (GCompareFunc)strcmp);
-  if (!iter)
-    return;
-  g_free (iter->data);
-  system->priv->cached_favorites = g_list_delete_link (system->priv->cached_favorites, iter);
-
-  val = gconf_value_new (GCONF_VALUE_LIST);
-  gconf_value_set_list_type (val, GCONF_VALUE_STRING);
-
-  set_gconf_value_string_list (val, system->priv->cached_favorites);
-  gconf_client_set (client, SHELL_APP_FAVORITES_KEY, val, NULL);
+  g_hash_table_remove (ref->appsys->priv->app_id_to_app, shell_app_info_get_id (ref->info));
+  shell_app_info_unref (ref->info);
+  g_free (ref);
 }
 
 /**
- * shell_app_system_lookup_app:
+ * shell_app_system_get_app:
  *
- * Return value: (transfer full): The #ShellAppInfo for id, or %NULL if none
+ * Find or create a #ShellApp corresponding to an id; if already cached
+ * elsewhere in memory, return that instance.  Otherwise, create a new
+ * one.
+ *
+ * Return value: (transfer full): The #ShellApp for id, or %NULL if none
  */
-ShellAppInfo *
-shell_app_system_lookup_cached_app (ShellAppSystem *self, const char *id)
+ShellApp *
+shell_app_system_get_app (ShellAppSystem   *self,
+                          const char       *id)
 {
   ShellAppInfo *info;
+  ShellApp *app;
 
-  info = g_hash_table_lookup (self->priv->app_id_to_app, id);
-  if (info)
-    shell_app_info_ref (info);
-  return info;
+  app = g_hash_table_lookup (self->priv->app_id_to_app, id);
+  if (app)
+    return g_object_ref (app);
+
+  info = g_hash_table_lookup (self->priv->app_id_to_info, id);
+  if (!info)
+    return NULL;
+
+  app = _shell_app_new (info);
+
+  return app;
+}
+
+/**
+ * shell_app_system_get_app_for_window:
+ * @self: A #ShellAppSystem
+ * @window: A #MetaWindow
+ *
+ * Find or create a #ShellApp for window
+ *
+ * Return value: (transfer full): The #ShellApp for window, or %NULL if none
+ */
+ShellApp *
+shell_app_system_get_app_for_window (ShellAppSystem *self,
+                                     MetaWindow *window)
+{
+  char *id = g_strdup_printf ("window:%p", window);
+  ShellApp *app = g_hash_table_lookup (self->priv->app_id_to_app, id);
+
+  if (app)
+    g_object_ref (G_OBJECT (app));
+  else
+    app = _shell_app_new_for_window (window);
+
+  g_free (id);
+
+  return app;
+}
+
+/* ShellAppSystem ensures we have a unique instance of
+ * apps per id.
+ */
+void
+_shell_app_system_register_app (ShellAppSystem   *self,
+                                ShellApp         *app)
+{
+  const char *id;
+  ShellAppRef *ref;
+
+  id = shell_app_get_id (app);
+
+  g_return_if_fail (g_hash_table_lookup (self->priv->app_id_to_app, id) == NULL);
+
+  ref = g_new0 (ShellAppRef, 1);
+  ref->appsys = self;
+  ref->info = shell_app_info_ref (_shell_app_get_info (app));
+  g_hash_table_insert (self->priv->app_id_to_app, (char*)shell_app_info_get_id (ref->info), app);
+  g_object_weak_ref (G_OBJECT (app), shell_app_system_on_app_weakref, ref);
 }
 
 ShellAppInfo *
@@ -752,16 +592,16 @@ shell_app_system_create_from_window (ShellAppSystem *system, MetaWindow *window)
  * heuristically determined application identifier
  * string, or %NULL if none.
  *
- * Returns: (transfer full): A #ShellAppInfo for name
+ * Returns: (transfer full): A #ShellApp for name
  */
-ShellAppInfo *
+ShellApp *
 shell_app_system_lookup_heuristic_basename (ShellAppSystem *system,
                                             const char *name)
 {
-  ShellAppInfo *result;
+  ShellApp *result;
   char **vendor_prefixes;
 
-  result = shell_app_system_lookup_cached_app (system, name);
+  result = shell_app_system_get_app (system, name);
   if (result != NULL)
     return result;
 
@@ -769,7 +609,7 @@ shell_app_system_lookup_heuristic_basename (ShellAppSystem *system,
        *vendor_prefixes; vendor_prefixes++)
     {
       char *tmpid = g_strjoin (NULL, *vendor_prefixes, "-", name, NULL);
-      result = shell_app_system_lookup_cached_app (system, tmpid);
+      result = shell_app_system_get_app (system, tmpid);
       g_free (tmpid);
       if (result != NULL)
         return result;
@@ -778,6 +618,287 @@ shell_app_system_lookup_heuristic_basename (ShellAppSystem *system,
   return NULL;
 }
 
+typedef enum {
+  MATCH_NONE,
+  MATCH_MULTIPLE, /* Matches multiple terms */
+  MATCH_PREFIX, /* Strict prefix */
+  MATCH_SUBSTRING /* Not prefix, substring */
+} ShellAppInfoSearchMatch;
+
+static char *
+normalize_and_casefold (const char *str)
+{
+  char *normalized, *result;
+
+  if (str == NULL)
+    return NULL;
+
+  normalized = g_utf8_normalize (str, -1, G_NORMALIZE_ALL);
+  result = g_utf8_casefold (normalized, -1);
+  g_free (normalized);
+  return result;
+}
+
+static char *
+trim_exec_line (const char *str)
+{
+  const char *start, *end, *pos;
+
+  end = strchr (str, ' ');
+  if (end == NULL)
+    end = str + strlen (str);
+
+  start = str;
+  while ((pos = strchr (start, '/')) && pos < end)
+    start = ++pos;
+
+  return g_strndup (start, end - start);
+}
+
+static void
+shell_app_info_init_search_data (ShellAppInfo *info)
+{
+  const char *name;
+  const char *exec;
+  const char *comment;
+  char *normalized_exec;
+
+  g_assert (info->type == SHELL_APP_INFO_TYPE_ENTRY);
+
+  name = gmenu_tree_entry_get_name ((GMenuTreeEntry*)info->entry);
+  info->casefolded_name = normalize_and_casefold (name);
+
+  comment = gmenu_tree_entry_get_comment ((GMenuTreeEntry*)info->entry);
+  info->casefolded_description = normalize_and_casefold (comment);
+
+  exec = gmenu_tree_entry_get_exec ((GMenuTreeEntry*)info->entry);
+  normalized_exec = normalize_and_casefold (exec);
+  info->casefolded_exec = trim_exec_line (normalized_exec);
+  g_free (normalized_exec);
+}
+
+static ShellAppInfoSearchMatch
+shell_app_info_match_terms (ShellAppInfo  *info,
+                            GSList        *terms)
+{
+  GSList *iter;
+  ShellAppInfoSearchMatch match;
+
+  if (G_UNLIKELY(!info->casefolded_name))
+    shell_app_info_init_search_data (info);
+
+  match = MATCH_NONE;
+  for (iter = terms; iter; iter = iter->next)
+    {
+      const char *term = iter->data;
+      const char *p;
+
+      p = strstr (info->casefolded_name, term);
+      if (p == info->casefolded_name)
+        {
+          if (match != MATCH_NONE)
+            return MATCH_MULTIPLE;
+          else
+            match = MATCH_PREFIX;
+         }
+      else if (p != NULL)
+        match = MATCH_SUBSTRING;
+
+      p = strstr (info->casefolded_exec, term);
+      if (p == info->casefolded_exec)
+        {
+          if (match != MATCH_NONE)
+            return MATCH_MULTIPLE;
+          else
+            match = MATCH_PREFIX;
+         }
+      else if (p != NULL)
+        match = MATCH_SUBSTRING;
+
+      if (!info->casefolded_description)
+        continue;
+      p = strstr (info->casefolded_description, term);
+      if (p != NULL)
+        match = MATCH_SUBSTRING;
+    }
+  return match;
+}
+
+static gint
+shell_app_info_compare (gconstpointer a,
+                        gconstpointer b,
+                        gpointer      data)
+{
+  ShellAppSystem *system = data;
+  const char *id_a = a;
+  const char *id_b = b;
+  ShellAppInfo *info_a = g_hash_table_lookup (system->priv->app_id_to_info, id_a);
+  ShellAppInfo *info_b = g_hash_table_lookup (system->priv->app_id_to_info, id_b);
+
+  if (!info_a->name_collation_key)
+    info_a->name_collation_key = g_utf8_collate_key (gmenu_tree_entry_get_name ((GMenuTreeEntry*)info_a->entry), -1);
+  if (!info_b->name_collation_key)
+    info_b->name_collation_key = g_utf8_collate_key (gmenu_tree_entry_get_name ((GMenuTreeEntry*)info_b->entry), -1);
+
+  return strcmp (info_a->name_collation_key, info_b->name_collation_key);
+}
+
+static GSList *
+sort_and_concat_results (ShellAppSystem *system,
+                         GSList         *multiple_matches,
+                         GSList         *prefix_matches,
+                         GSList         *substring_matches)
+{
+  multiple_matches = g_slist_sort_with_data (multiple_matches, shell_app_info_compare, system);
+  prefix_matches = g_slist_sort_with_data (prefix_matches, shell_app_info_compare, system);
+  substring_matches = g_slist_sort_with_data (substring_matches, shell_app_info_compare, system);
+  return g_slist_concat (multiple_matches, g_slist_concat (prefix_matches, substring_matches));
+}
+
+/**
+ * normalize_terms:
+ * @terms: (element-type utf8): Input search terms
+ *
+ * Returns: (element-type utf8) (transfer full): Unicode-normalized and lowercased terms
+ */
+static GSList *
+normalize_terms (GSList *terms)
+{
+  GSList *normalized_terms = NULL;
+  GSList *iter;
+  for (iter = terms; iter; iter = iter->next)
+    {
+      const char *term = iter->data;
+      normalized_terms = g_slist_prepend (normalized_terms, normalize_and_casefold (term));
+    }
+  return normalized_terms;
+}
+
+static inline void
+shell_app_system_do_match (ShellAppSystem   *system,
+                           ShellAppInfo     *info,
+                           GSList           *terms,
+                           GSList          **multiple_results,
+                           GSList          **prefix_results,
+                           GSList          **substring_results)
+{
+  const char *id = shell_app_info_get_id (info);
+  ShellAppInfoSearchMatch match;
+
+  if (shell_app_info_get_is_nodisplay (info))
+    return;
+
+  match = shell_app_info_match_terms (info, terms);
+  switch (match)
+    {
+      case MATCH_NONE:
+        break;
+      case MATCH_MULTIPLE:
+        *multiple_results = g_slist_prepend (*multiple_results, (char *) id);
+        break;
+      case MATCH_PREFIX:
+        *prefix_results = g_slist_prepend (*prefix_results, (char *) id);
+        break;
+      case MATCH_SUBSTRING:
+        *substring_results = g_slist_prepend (*substring_results, (char *) id);
+        break;
+    }
+}
+
+static GSList *
+shell_app_system_initial_search_internal (ShellAppSystem  *self,
+                                          GSList          *terms,
+                                          GSList          *source)
+{
+  GSList *multiple_results = NULL;
+  GSList *prefix_results = NULL;
+  GSList *substring_results = NULL;
+  GSList *iter;
+  GSList *normalized_terms = normalize_terms (terms);
+
+  for (iter = source; iter; iter = iter->next)
+    {
+      ShellAppInfo *info = iter->data;
+
+      shell_app_system_do_match (self, info, normalized_terms, &multiple_results, &prefix_results, &substring_results);
+    }
+  g_slist_foreach (normalized_terms, (GFunc)g_free, NULL);
+  g_slist_free (normalized_terms);
+
+  return sort_and_concat_results (self, multiple_results, prefix_results, substring_results);
+}
+
+/**
+ * shell_app_system_initial_search:
+ * @self: A #ShellAppSystem
+ * @prefs: %TRUE if we should search preferences instead of apps
+ * @terms: (element-type utf8): List of terms, logical OR
+ *
+ * Search through applications for the given search terms.  Note that returned
+ * strings are only valid until a return to the main loop.
+ *
+ * Returns: (transfer container) (element-type utf8): List of application identifiers
+ */
+GSList *
+shell_app_system_initial_search (ShellAppSystem  *self,
+                                 gboolean         prefs,
+                                 GSList          *terms)
+{
+  return shell_app_system_initial_search_internal (self, terms,
+            prefs ? self->priv->cached_settings : self->priv->cached_flattened_apps);
+}
+
+/**
+ * shell_app_system_subsearch:
+ * @self: A #ShellAppSystem
+ * @prefs: %TRUE if we should search preferences instead of apps
+ * @previous_results: (element-type utf8): List of previous results
+ * @terms: (element-type utf8): List of terms, logical OR
+ *
+ * Search through a previous result set; for more information, see
+ * js/ui/search.js.  Note the value of @prefs must be
+ * the same as passed to shell_app_system_initial_search().  Note that returned
+ * strings are only valid until a return to the main loop.
+ *
+ * Returns: (transfer container) (element-type utf8): List of application identifiers
+ */
+GSList *
+shell_app_system_subsearch (ShellAppSystem   *system,
+                            gboolean          prefs,
+                            GSList           *previous_results,
+                            GSList           *terms)
+{
+  GSList *iter;
+  GSList *multiple_results = NULL;
+  GSList *prefix_results = NULL;
+  GSList *substring_results = NULL;
+  GSList *normalized_terms = normalize_terms (terms);
+
+  /* Note prefs is deliberately ignored; both apps and prefs are in app_id_to_app,
+   * but we have the parameter for consistency and in case in the future
+   * they're not in the same data structure.
+   */
+
+  for (iter = previous_results; iter; iter = iter->next)
+    {
+      const char *id = iter->data;
+      ShellAppInfo *info;
+
+      info = g_hash_table_lookup (system->priv->app_id_to_info, id);
+      if (!info)
+        continue;
+
+      shell_app_system_do_match (system, info, normalized_terms, &multiple_results, &prefix_results, &substring_results);
+    }
+  g_slist_foreach (normalized_terms, (GFunc)g_free, NULL);
+  g_slist_free (normalized_terms);
+
+  /* Note that a shorter term might have matched as a prefix, but
+     when extended only as a substring, so we have to redo the
+     sort rather than reusing the existing ordering */
+  return sort_and_concat_results (system, multiple_results, prefix_results, substring_results);
+}
+
 const char *
 shell_app_info_get_id (ShellAppInfo *info)
 {
@@ -809,6 +930,8 @@ shell_app_info_get_name (ShellAppInfo *info)
       {
         char *title;
         g_object_get (info->window, "title", &title, NULL);
+        if (!title)
+          title = g_strdup ("");
         return title;
       }
   }
@@ -898,7 +1021,16 @@ themed_icon_from_name (const char *iconname)
   return icon;
 }
 
-static GIcon *
+/**
+ * shell_app_info_get_icon:
+ * @info: A #ShellAppInfo
+ *
+ * Get the #GIcon associated with this app; for apps "faked" from a #MetaWindow,
+ * return %NULL.
+ *
+ * Returns: (transfer full): The icon for @info, or %NULL
+ */
+GIcon *
 shell_app_info_get_icon (ShellAppInfo *info)
 {
   char *iconname = NULL;
@@ -977,20 +1109,39 @@ shell_app_info_create_icon_texture (ShellAppInfo *info, float size)
 
   if (info->type == SHELL_APP_INFO_TYPE_WINDOW)
     {
-      return shell_texture_cache_bind_pixbuf_property (shell_texture_cache_get_default (),
-                                                       G_OBJECT (info->window),
-                                                       "icon");
+      return st_texture_cache_bind_pixbuf_property (st_texture_cache_get_default (),
+                                                    G_OBJECT (info->window),
+                                                    "icon");
     }
 
   icon = shell_app_info_get_icon (info);
-  if (!icon)
+  if (icon == NULL)
     {
       ret = clutter_texture_new ();
       g_object_set (ret, "opacity", 0, "width", size, "height", size, NULL);
-      return ret;
+    }
+  else
+    {
+      ret = st_texture_cache_load_gicon (st_texture_cache_get_default (), icon, (int)size);
+      g_object_unref (icon);
     }
 
-  return shell_texture_cache_load_gicon (shell_texture_cache_get_default (), icon, (int)size);
+  return ret;
+}
+
+/**
+ * shell_app_info_get_source_window:
+ * @info: A #ShellAppInfo
+ *
+ * If @info is tracking a #MetaWindow, return that window.
+ * Otherwise, return %NULL.
+ */
+MetaWindow *
+shell_app_info_get_source_window (ShellAppInfo *info)
+{
+  if (info->type == SHELL_APP_INFO_TYPE_WINDOW)
+    return info->window;
+  return NULL;
 }
 
 /**
