@@ -32,11 +32,41 @@
  * allow the user to move around the scrollable area.
  */
 
+/* TODO: The code here currently only deals with height-for-width
+ * allocation; width-for-height allocation would need a second set of
+ * code paths through get_preferred_height()/get_preferred_width()/allocate()
+ * that reverse the roles of the horizontal and vertical scrollbars.
+ *
+ * TODO: The multiple layout passes with and without scrollbars when
+ * using the automatic policy causes considerable inefficiency because
+ * it breaks request caching; we should saved the last size passed
+ * into allocate() and if it's the same as previous size not repeat
+ * the determination of scrollbar visibility. This requires overriding
+ * queue_relayout() so we know when to discard the saved value.
+ *
+ * The size negotiation between the #StScrollView and the child is
+ * described in the documentation for #StScrollable; the significant
+ * part to note there is that reported minimum sizes for a scrolled
+ * child are the minimum sizes when no scrollbar is needed. This allows
+ * us to determine what scrollbars are visible without a need to look
+ * inside the #StAdjustment.
+ *
+ * The second simplification that we make that allows us to implement
+ * a straighforward height-for-width negotiation without multiple
+ * allocate passes is that when the vertical scrollbar policy is
+ * AUTO, we always reserve space for the vertical scrollbar in the
+ * reported minimum and natural size.
+ *
+ * See https://bugzilla.gnome.org/show_bug.cgi?id=611740 for a more
+ * detailed description of the considerations involved.
+ */
+
 #include "st-scroll-view.h"
 #include "st-marshal.h"
 #include "st-scroll-bar.h"
 #include "st-scrollable.h"
 #include <clutter/clutter.h>
+#include <math.h>
 
 static void clutter_container_iface_init (ClutterContainerIface *iface);
 
@@ -61,15 +91,28 @@ struct _StScrollViewPrivate
    */
   ClutterActor *child;
 
+  StAdjustment *hadjustment;
   ClutterActor *hscroll;
+  StAdjustment *vadjustment;
   ClutterActor *vscroll;
 
+  GtkPolicyType hscrollbar_policy;
+  GtkPolicyType vscrollbar_policy;
+
+  ClutterActor *top_shadow;
+  ClutterActor *bottom_shadow;
+
   gfloat        row_size;
   gfloat        column_size;
 
+  gboolean      vshadows;
   gboolean      row_size_set : 1;
   gboolean      column_size_set : 1;
   guint         mouse_scroll : 1;
+  guint         hscrollbar_visible : 1;
+  guint         vscrollbar_visible : 1;
+  guint         top_shadow_visible : 1;
+  guint         bottom_shadow_visible : 1;
 };
 
 enum {
@@ -77,7 +120,10 @@ enum {
 
   PROP_HSCROLL,
   PROP_VSCROLL,
-  PROP_MOUSE_SCROLL
+  PROP_HSCROLLBAR_POLICY,
+  PROP_VSCROLLBAR_POLICY,
+  PROP_MOUSE_SCROLL,
+  PROP_VSHADOWS
 };
 
 static void
@@ -96,25 +142,119 @@ st_scroll_view_get_property (GObject    *object,
     case PROP_VSCROLL:
       g_value_set_object (value, priv->vscroll);
       break;
+    case PROP_HSCROLLBAR_POLICY:
+      g_value_set_enum (value, priv->hscrollbar_policy);
+      break;
+    case PROP_VSCROLLBAR_POLICY:
+      g_value_set_enum (value, priv->vscrollbar_policy);
+      break;
     case PROP_MOUSE_SCROLL:
       g_value_set_boolean (value, priv->mouse_scroll);
       break;
+    case PROP_VSHADOWS:
+      g_value_set_boolean (value, priv->vshadows);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
 }
 
+static void
+update_shadow_visibility (StScrollView *scroll)
+{
+  StScrollViewPrivate *priv = scroll->priv;
+
+  if (priv->vshadows)
+    {
+      gdouble value, lower, upper, page_size;
+
+      st_adjustment_get_values (priv->vadjustment, &value, &lower, &upper, NULL, NULL, &page_size);
+
+      priv->top_shadow_visible = value > lower + 0.1;
+      priv->bottom_shadow_visible = value < upper - page_size - 0.1;
+    }
+  else
+    {
+      priv->top_shadow_visible = FALSE;
+      priv->bottom_shadow_visible = FALSE;
+    }
+}
+
+/**
+ * st_scroll_view_set_vshadows:
+ * @self: a #StScrollView
+ * @vshadows: Whether to enable vertical shadows
+ *
+ * Sets whether to show shadows at the top and bottom of the area. Shadows
+ * are omitted when fully scrolled to that edge.
+ */
+void
+st_scroll_view_set_vshadows (StScrollView *self,
+                             gboolean      vshadows)
+{
+  StScrollViewPrivate *priv = ST_SCROLL_VIEW (self)->priv;
+
+  vshadows = vshadows != FALSE;
+  if (priv->vshadows == vshadows)
+    return;
+
+  priv->vshadows = vshadows;
+
+  if (vshadows)
+    {
+      if (priv->top_shadow)
+        {
+          clutter_actor_show (priv->top_shadow);
+          clutter_actor_show (priv->bottom_shadow);
+        }
+      else
+        {
+          priv->top_shadow = g_object_new (ST_TYPE_BIN, "style-class", "top-shadow", NULL);
+          priv->bottom_shadow = g_object_new (ST_TYPE_BIN, "style-class", "bottom-shadow", NULL);
+
+          clutter_actor_set_parent (priv->bottom_shadow, CLUTTER_ACTOR (self));
+          clutter_actor_set_parent (priv->top_shadow, CLUTTER_ACTOR (self));
+        }
+    }
+  else
+    {
+      clutter_actor_hide (priv->top_shadow);
+      clutter_actor_hide (priv->bottom_shadow);
+    }
+
+  update_shadow_visibility (self);
+
+  g_object_notify (G_OBJECT (self), "vshadows");
+}
+
 static void
 st_scroll_view_set_property (GObject      *object,
                              guint         property_id,
                              const GValue *value,
                              GParamSpec   *pspec)
 {
+  StScrollView *self = ST_SCROLL_VIEW (object);
+  StScrollViewPrivate *priv = self->priv;
+
   switch (property_id)
     {
+    case PROP_VSHADOWS:
+      st_scroll_view_set_vshadows (self, g_value_get_boolean (value));
+      break;
     case PROP_MOUSE_SCROLL:
-      st_scroll_view_set_mouse_scrolling ((StScrollView *) object,
+      st_scroll_view_set_mouse_scrolling (self,
                                           g_value_get_boolean (value));
+      break;
+    case PROP_HSCROLLBAR_POLICY:
+      st_scroll_view_set_policy (self,
+                                 g_value_get_enum (value),
+                                 priv->vscrollbar_policy);
+      break;
+    case PROP_VSCROLLBAR_POLICY:
+      st_scroll_view_set_policy (self,
+                                 priv->hscrollbar_policy,
+                                 g_value_get_enum (value));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -125,29 +265,48 @@ st_scroll_view_dispose (GObject *object)
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (object)->priv;
 
-  priv->child = NULL;
-
   if (priv->vscroll)
-    {
-      clutter_actor_unparent (priv->vscroll);
-      priv->vscroll = NULL;
-    }
+    clutter_actor_destroy (priv->vscroll);
 
   if (priv->hscroll)
+    clutter_actor_destroy (priv->hscroll);
+
+  /* For most reliable freeing of memory, an object with signals
+   * like StAdjustment should be explicitly disposed. Since we own
+   * the adjustments, we take care of that. This also disconnects
+   * the signal handlers that we established on creation.
+   */
+  if (priv->hadjustment)
     {
-      clutter_actor_unparent (priv->hscroll);
-      priv->hscroll = NULL;
+      g_object_run_dispose (G_OBJECT (priv->hadjustment));
+      g_object_unref (priv->hadjustment);
+      priv->hadjustment = NULL;
+    }
+
+  if (priv->vadjustment)
+    {
+      g_object_run_dispose (G_OBJECT (priv->vadjustment));
+      g_object_unref (priv->vadjustment);
+      priv->hadjustment = NULL;
+    }
+
+  /* since it's impossible to get a handle to these actors, we can
+   * just directly unparent them and not go through destroy/remove */
+  if (priv->top_shadow)
+    {
+      clutter_actor_unparent (priv->top_shadow);
+      priv->top_shadow = NULL;
+    }
+
+  if (priv->bottom_shadow)
+    {
+      clutter_actor_unparent (priv->bottom_shadow);
+      priv->bottom_shadow = NULL;
     }
 
   G_OBJECT_CLASS (st_scroll_view_parent_class)->dispose (object);
 }
 
-static void
-st_scroll_view_finalize (GObject *object)
-{
-  G_OBJECT_CLASS (st_scroll_view_parent_class)->finalize (object);
-}
-
 static void
 st_scroll_view_paint (ClutterActor *actor)
 {
@@ -157,10 +316,15 @@ st_scroll_view_paint (ClutterActor *actor)
   CLUTTER_ACTOR_CLASS (st_scroll_view_parent_class)->paint (actor);
 
   /* paint our custom children */
-  if (CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
+  if (priv->hscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
     clutter_actor_paint (priv->hscroll);
-  if (CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
+  if (priv->vscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
     clutter_actor_paint (priv->vscroll);
+
+  if (priv->top_shadow_visible)
+    clutter_actor_paint (priv->top_shadow);
+  if (priv->bottom_shadow_visible)
+    clutter_actor_paint (priv->bottom_shadow);
 }
 
 static void
@@ -173,32 +337,46 @@ st_scroll_view_pick (ClutterActor       *actor,
   CLUTTER_ACTOR_CLASS (st_scroll_view_parent_class)->pick (actor, color);
 
   /* paint our custom children */
-  if (CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
+  if (priv->hscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
     clutter_actor_paint (priv->hscroll);
-  if (CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
+  if (priv->vscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
     clutter_actor_paint (priv->vscroll);
 }
 
 static double
-get_scrollbar_width (StScrollView *scroll_view)
+get_scrollbar_width (StScrollView *scroll)
 {
-  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (scroll_view));
-  double result = DEFAULT_SCROLLBAR_WIDTH;
+  StScrollViewPrivate *priv = scroll->priv;
 
-  st_theme_node_get_length (theme_node, "scrollbar-width", FALSE, &result);
+  if (CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
+    {
+      StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (scroll));
+      double result = DEFAULT_SCROLLBAR_WIDTH;
 
-  return result;
+      st_theme_node_get_length (theme_node, "scrollbar-width", FALSE, &result);
+
+      return result;
+    }
+  else
+    return 0;
 }
 
 static double
-get_scrollbar_height (StScrollView *scroll_view)
+get_scrollbar_height (StScrollView *scroll)
 {
-  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (scroll_view));
-  double result = DEFAULT_SCROLLBAR_WIDTH;
+  StScrollViewPrivate *priv = scroll->priv;
 
-  st_theme_node_get_length (theme_node, "scrollbar-height", FALSE, &result);
+  if (CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
+    {
+      StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (scroll));
+      double result = DEFAULT_SCROLLBAR_WIDTH;
 
-  return result;
+      st_theme_node_get_length (theme_node, "scrollbar-height", FALSE, &result);
+
+      return result;
+    }
+  else
+    return 0;
 }
 
 static void
@@ -209,32 +387,64 @@ st_scroll_view_get_preferred_width (ClutterActor *actor,
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (actor)->priv;
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
+  gboolean account_for_vscrollbar = FALSE;
+  gfloat min_width = 0, natural_width;
+  gfloat child_min_width, child_natural_width;
 
   if (!priv->child)
     return;
 
   st_theme_node_adjust_for_height (theme_node, &for_height);
 
-  /* Our natural width is the natural width of the child */
-  clutter_actor_get_preferred_width (priv->child,
-                                     for_height,
-                                     NULL,
-                                     natural_width_p);
+  clutter_actor_get_preferred_width (priv->child, -1,
+                                     &child_min_width, &child_natural_width);
 
-  /* Add space for the scroll-bar if we can determine it will be necessary */
-  if ((for_height >= 0) && natural_width_p)
+  natural_width = child_natural_width;
+
+  switch (priv->hscrollbar_policy)
     {
-      gfloat natural_height;
+    case GTK_POLICY_NEVER:
+      min_width = child_min_width;
+      break;
+    case GTK_POLICY_ALWAYS:
+    case GTK_POLICY_AUTOMATIC:
+      /* Should theoretically use the min width of the hscrollbar,
+       * but that's not cleanly defined at the moment */
+      min_width = 0;
+      break;
+    }
 
-      clutter_actor_get_preferred_height (priv->child, -1.0,
-                                          NULL,
-                                          &natural_height);
-      if (for_height < natural_height)
-        *natural_width_p += get_scrollbar_width (ST_SCROLL_VIEW (actor));
+  switch (priv->vscrollbar_policy)
+    {
+    case GTK_POLICY_NEVER:
+      account_for_vscrollbar = FALSE;
+      break;
+    case GTK_POLICY_ALWAYS:
+      account_for_vscrollbar = TRUE;
+      break;
+    case GTK_POLICY_AUTOMATIC:
+      /* For automatic scrollbars, we always request space for the vertical
+       * scrollbar; we won't know whether we actually need one until our
+       * height is assigned in allocate().
+       */
+      account_for_vscrollbar = TRUE;
+      break;
+    }
+
+  account_for_vscrollbar = priv->vscrollbar_policy != GTK_POLICY_NEVER;
+  if (account_for_vscrollbar)
+    {
+      float sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor));
+
+      min_width += sb_width;
+      natural_width += sb_width;
     }
 
   if (min_width_p)
-    *min_width_p = 0;
+    *min_width_p = min_width;
+
+  if (natural_width_p)
+    *natural_width_p = natural_width;
 
   st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
 }
@@ -247,36 +457,98 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (actor)->priv;
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
+  gboolean account_for_hscrollbar = FALSE;
+  gfloat min_height = 0, natural_height;
+  gfloat child_min_height, child_natural_height;
+  gfloat child_min_width;
+  gfloat sb_width;
 
   if (!priv->child)
     return;
 
   st_theme_node_adjust_for_width (theme_node, &for_width);
 
-  /* Our natural height is the natural height of the child */
-  clutter_actor_get_preferred_height (priv->child,
-                                      for_width,
-                                      NULL,
-                                      natural_height_p);
+  clutter_actor_get_preferred_width (priv->child, -1,
+                                     &child_min_width, NULL);
 
-  /* Add space for the scroll-bar if we can determine it will be necessary */
-  if ((for_width >= 0) && natural_height_p)
-    {
-      gfloat natural_width;
-
-      clutter_actor_get_preferred_width (priv->child, -1.0,
-                                         NULL,
-                                         &natural_width);
-      if (for_width < natural_width)
-        *natural_height_p += get_scrollbar_height (ST_SCROLL_VIEW (actor));
-    }
+  natural_height = child_natural_height;
 
   if (min_height_p)
     *min_height_p = 0;
 
+  sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor));
+
+  switch (priv->vscrollbar_policy)
+    {
+    case GTK_POLICY_NEVER:
+      break;
+    case GTK_POLICY_ALWAYS:
+    case GTK_POLICY_AUTOMATIC:
+      /* We've requested space for the scrollbar, subtract it back out */
+      for_width -= sb_width;
+      break;
+    }
+
+  switch (priv->hscrollbar_policy)
+    {
+    case GTK_POLICY_NEVER:
+      account_for_hscrollbar = FALSE;
+      break;
+    case GTK_POLICY_ALWAYS:
+      account_for_hscrollbar = TRUE;
+      break;
+    case GTK_POLICY_AUTOMATIC:
+      account_for_hscrollbar = for_width < child_min_width;
+      break;
+    }
+
+  clutter_actor_get_preferred_height (priv->child, for_width,
+                                      &child_min_height, &child_natural_height);
+
+  natural_height = child_natural_height;
+
+  switch (priv->vscrollbar_policy)
+    {
+    case GTK_POLICY_NEVER:
+      min_height = child_min_height;
+      break;
+    case GTK_POLICY_ALWAYS:
+    case GTK_POLICY_AUTOMATIC:
+      /* Should theoretically use the min height of the vscrollbar,
+       * but that's not cleanly defined at the moment */
+      min_height = 0;
+      break;
+    }
+
+  if (account_for_hscrollbar)
+    {
+      float sb_height = get_scrollbar_height (ST_SCROLL_VIEW (actor));
+
+      min_height += sb_height;
+      natural_height += sb_height;
+    }
+
+  if (min_height_p)
+    *min_height_p = min_height;
+
+  if (natural_height_p)
+    *natural_height_p = natural_height;
+
   st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
 }
 
+static gfloat
+get_shadow_height (ClutterActor *shadow)
+{
+  gfloat natural_height;
+
+  /* The shadows are empty StBin and have no height-for-width behavior */
+
+  clutter_actor_get_preferred_height (shadow, -1, NULL, &natural_height);
+
+  return natural_height;
+}
+
 static void
 st_scroll_view_allocate (ClutterActor          *actor,
                          const ClutterActorBox *box,
@@ -285,6 +557,7 @@ st_scroll_view_allocate (ClutterActor          *actor,
   ClutterActorBox content_box, child_box;
   ClutterActorClass *parent_parent_class;
   gfloat avail_width, avail_height, sb_width, sb_height;
+  gboolean hscrollbar_visible, vscrollbar_visible;
 
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (actor)->priv;
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
@@ -308,13 +581,71 @@ st_scroll_view_allocate (ClutterActor          *actor,
   avail_height = content_box.y2 - content_box.y1;
 
   sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor));
-  sb_height = get_scrollbar_width (ST_SCROLL_VIEW (actor));
+  sb_height = get_scrollbar_height (ST_SCROLL_VIEW (actor));
 
-  if (!CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
-    sb_width = 0;
+  /* Determine what scrollbars are visible. The basic idea of the
+   * handling of an automatic scrollbars is that we start off with the
+   * assumption that we don't need any scrollbars, see if that works,
+   * and if not add horizontal and vertical scrollbars until we are no
+   * longer overflowing.
+   */
+  if (priv->child)
+    {
+      gfloat child_min_width;
+      gfloat child_min_height;
 
-  if (!CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
-    sb_height = 0;
+      clutter_actor_get_preferred_width (priv->child, -1,
+                                         &child_min_width, NULL);
+
+      if (priv->vscrollbar_policy == GTK_POLICY_AUTOMATIC)
+        {
+          if (priv->hscrollbar_policy == GTK_POLICY_AUTOMATIC)
+            {
+              /* Pass one, try without a vertical scrollbar */
+              clutter_actor_get_preferred_height (priv->child, avail_width, &child_min_height, NULL);
+              vscrollbar_visible = child_min_height > avail_height;
+              hscrollbar_visible = child_min_width > avail_width - (vscrollbar_visible ? sb_width : 0);
+              vscrollbar_visible = child_min_height > avail_height - (hscrollbar_visible ? sb_height : 0);
+
+              /* Pass two - if we needed a vertical scrollbar, get a new preferred height */
+              if (vscrollbar_visible)
+                {
+                  clutter_actor_get_preferred_height (priv->child, MAX (avail_width - sb_width, 0),
+                                                      &child_min_height, NULL);
+                  hscrollbar_visible = child_min_width > avail_width - sb_width;
+                }
+            }
+          else
+            {
+              hscrollbar_visible = priv->hscrollbar_policy != GTK_POLICY_NEVER;
+
+              /* try without a vertical scrollbar */
+              clutter_actor_get_preferred_height (priv->child, avail_width, &child_min_height, NULL);
+              vscrollbar_visible = child_min_height > avail_height - (hscrollbar_visible ? sb_height : 0);
+            }
+        }
+      else
+        {
+          vscrollbar_visible = priv->vscrollbar_policy != GTK_POLICY_NEVER;
+
+          if (priv->hscrollbar_policy == GTK_POLICY_AUTOMATIC)
+            hscrollbar_visible = child_min_width > avail_height - (vscrollbar_visible ? 0 : sb_width);
+          else
+            hscrollbar_visible = priv->hscrollbar_policy != GTK_POLICY_NEVER;
+        }
+    }
+  else
+    {
+      hscrollbar_visible = priv->hscrollbar_policy != GTK_POLICY_NEVER;
+      vscrollbar_visible = priv->vscrollbar_policy != GTK_POLICY_NEVER;
+    }
+
+  /* Whether or not we show the scrollbars, if the scrollbars are visible
+   * actors, we need to give them some allocation, so we unconditionally
+   * give them the "right" allocation; that might overlap the child when
+   * the scrollbars are not visible, but it doesn't matter because we
+   * don't include them in pick or paint.
+   */
 
   /* Vertical scrollbar */
   if (CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
@@ -338,6 +669,11 @@ st_scroll_view_allocate (ClutterActor          *actor,
       clutter_actor_allocate (priv->hscroll, &child_box, flags);
     }
 
+  /* Fold visibility into the scrollbar sizes to simplify the rest of the computations */
+  if (!hscrollbar_visible)
+    sb_height = 0;
+  if (!vscrollbar_visible)
+    sb_width = 0;
 
   /* Child */
   child_box.x1 = content_box.x1;
@@ -347,6 +683,28 @@ st_scroll_view_allocate (ClutterActor          *actor,
 
   if (priv->child)
     clutter_actor_allocate (priv->child, &child_box, flags);
+
+  /* Shadows */
+  if (priv->top_shadow && CLUTTER_ACTOR_IS_VISIBLE (priv->top_shadow))
+    {
+      child_box.x1 = content_box.x1;
+      child_box.y1 = content_box.y1;
+      child_box.x2 = MAX (child_box.x1, content_box.x2 - sb_width);
+      child_box.y2 = content_box.y1 + get_shadow_height (priv->top_shadow);
+      clutter_actor_allocate (priv->top_shadow, &child_box, flags);
+    }
+
+  if (priv->bottom_shadow && CLUTTER_ACTOR_IS_VISIBLE (priv->bottom_shadow))
+    {
+      child_box.x1 = content_box.x1;
+      child_box.y1 = content_box.y2 - sb_height - get_shadow_height (priv->bottom_shadow);
+      child_box.x2 = MAX (content_box.x1, content_box.x2 - sb_width);
+      child_box.y2 = content_box.y2 - sb_height;
+      clutter_actor_allocate (priv->bottom_shadow, &child_box, flags);
+    }
+
+  priv->hscrollbar_visible = hscrollbar_visible;
+  priv->vscrollbar_visible = vscrollbar_visible;
 }
 
 static void
@@ -357,6 +715,12 @@ st_scroll_view_style_changed (StWidget *widget)
   st_widget_style_changed (ST_WIDGET (priv->hscroll));
   st_widget_style_changed (ST_WIDGET (priv->vscroll));
 
+  if (priv->top_shadow)
+    {
+      st_widget_style_changed (ST_WIDGET (priv->top_shadow));
+      st_widget_style_changed (ST_WIDGET (priv->bottom_shadow));
+    }
+
   ST_WIDGET_CLASS (st_scroll_view_parent_class)->style_changed (widget);
 }
 
@@ -366,40 +730,30 @@ st_scroll_view_scroll_event (ClutterActor       *self,
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (self)->priv;
   gdouble lower, value, upper, step;
-  StAdjustment *vadjustment, *hadjustment;
 
   /* don't handle scroll events if requested not to */
   if (!priv->mouse_scroll)
     return FALSE;
 
-  hadjustment = st_scroll_bar_get_adjustment (ST_SCROLL_BAR(priv->hscroll));
-  vadjustment = st_scroll_bar_get_adjustment (ST_SCROLL_BAR(priv->vscroll));
-
   switch (event->direction)
     {
     case CLUTTER_SCROLL_UP:
     case CLUTTER_SCROLL_DOWN:
-      if (vadjustment)
-        g_object_get (vadjustment,
-                      "lower", &lower,
-                      "step-increment", &step,
-                      "value", &value,
-                      "upper", &upper,
-                      NULL);
-      else
-        return FALSE;
+      g_object_get (priv->vadjustment,
+                    "lower", &lower,
+                    "step-increment", &step,
+                    "value", &value,
+                    "upper", &upper,
+                    NULL);
       break;
     case CLUTTER_SCROLL_LEFT:
     case CLUTTER_SCROLL_RIGHT:
-      if (vadjustment)
-        g_object_get (hadjustment,
-                      "lower", &lower,
-                      "step-increment", &step,
-                      "value", &value,
-                      "upper", &upper,
-                      NULL);
-      else
-        return FALSE;
+      g_object_get (priv->hadjustment,
+                    "lower", &lower,
+                    "step-increment", &step,
+                    "value", &value,
+                    "upper", &upper,
+                    NULL);
       break;
     }
 
@@ -409,25 +763,25 @@ st_scroll_view_scroll_event (ClutterActor       *self,
       if (value == lower)
         return FALSE;
       else
-        st_adjustment_set_value (vadjustment, value - step);
+        st_adjustment_set_value (priv->vadjustment, value - step);
       break;
     case CLUTTER_SCROLL_DOWN:
       if (value == upper)
         return FALSE;
       else
-        st_adjustment_set_value (vadjustment, value + step);
+        st_adjustment_set_value (priv->vadjustment, value + step);
       break;
     case CLUTTER_SCROLL_LEFT:
       if (value == lower)
         return FALSE;
       else
-        st_adjustment_set_value (hadjustment, value - step);
+        st_adjustment_set_value (priv->hadjustment, value - step);
       break;
     case CLUTTER_SCROLL_RIGHT:
       if (value == upper)
         return FALSE;
       else
-        st_adjustment_set_value (hadjustment, value + step);
+        st_adjustment_set_value (priv->hadjustment, value + step);
       break;
     }
 
@@ -446,8 +800,7 @@ st_scroll_view_class_init (StScrollViewClass *klass)
 
   object_class->get_property = st_scroll_view_get_property;
   object_class->set_property = st_scroll_view_set_property;
-  object_class->dispose= st_scroll_view_dispose;
-  object_class->finalize = st_scroll_view_finalize;
+  object_class->dispose = st_scroll_view_dispose;
 
   actor_class->paint = st_scroll_view_paint;
   actor_class->pick = st_scroll_view_pick;
@@ -474,6 +827,23 @@ st_scroll_view_class_init (StScrollViewClass *klass)
                                                         ST_TYPE_SCROLL_BAR,
                                                         G_PARAM_READABLE));
 
+
+  pspec = g_param_spec_enum ("vscrollbar-policy",
+                             "Vertical Scrollbar Policy",
+                             "When the vertical scrollbar is displayed",
+                             GTK_TYPE_POLICY_TYPE,
+                             GTK_POLICY_AUTOMATIC,
+                             G_PARAM_READWRITE);
+  g_object_class_install_property (object_class, PROP_VSCROLLBAR_POLICY, pspec);
+
+  pspec = g_param_spec_enum ("hscrollbar-policy",
+                             "Horizontal Scrollbar Policy",
+                             "When the horizontal scrollbar is displayed",
+                             GTK_TYPE_POLICY_TYPE,
+                             GTK_POLICY_AUTOMATIC,
+                             G_PARAM_READWRITE);
+  g_object_class_install_property (object_class, PROP_HSCROLLBAR_POLICY, pspec);
+
   pspec = g_param_spec_boolean ("enable-mouse-scrolling",
                                 "Enable Mouse Scrolling",
                                 "Enable automatic mouse wheel scrolling",
@@ -483,98 +853,29 @@ st_scroll_view_class_init (StScrollViewClass *klass)
                                    PROP_MOUSE_SCROLL,
                                    pspec);
 
+  pspec = g_param_spec_boolean ("vshadows",
+                                "Vertical Shadows",
+                                "Show shadows at the top and and bottom of the area unless fully scrolled to that edge",
+                                FALSE,
+                                G_PARAM_READWRITE);
+  g_object_class_install_property (object_class,
+                                   PROP_VSHADOWS,
+                                   pspec);
 }
 
 static void
 child_adjustment_changed_cb (StAdjustment *adjustment,
-                             ClutterActor *bar)
+                             StScrollView *scroll)
 {
-  StScrollView *scroll;
-  gdouble lower, upper, page_size;
-
-  scroll = ST_SCROLL_VIEW (clutter_actor_get_parent (bar));
-
-  /* Determine if this scroll-bar should be visible */
-  st_adjustment_get_values (adjustment, NULL,
-                            &lower, &upper,
-                            NULL, NULL,
-                            &page_size);
-
-  if ((upper - lower) > page_size)
-    clutter_actor_show (bar);
-  else
-    clutter_actor_hide (bar);
-
-  /* Request a resize */
-  clutter_actor_queue_relayout (CLUTTER_ACTOR (scroll));
+  update_shadow_visibility (scroll);
 }
 
 static void
-child_hadjustment_notify_cb (GObject    *gobject,
-                             GParamSpec *arg1,
-                             gpointer    user_data)
+child_adjustment_notify_value (GObject      *gobject,
+                               GParamSpec   *pspec,
+                               StScrollView *scroll)
 {
-  StAdjustment *hadjust;
-
-  ClutterActor *actor = CLUTTER_ACTOR (gobject);
-  StScrollViewPrivate *priv = ST_SCROLL_VIEW (user_data)->priv;
-
-  hadjust = st_scroll_bar_get_adjustment (ST_SCROLL_BAR(priv->hscroll));
-  if (hadjust)
-    g_signal_handlers_disconnect_by_func (hadjust,
-                                          child_adjustment_changed_cb,
-                                          priv->hscroll);
-
-  st_scrollable_get_adjustments (ST_SCROLLABLE (actor), &hadjust, NULL);
-  if (hadjust)
-    {
-      /* Force scroll step if neede. */
-      if (priv->column_size_set)
-        {
-          g_object_set (hadjust,
-                        "step-increment", priv->column_size,
-                        NULL);
-        }
-
-      st_scroll_bar_set_adjustment (ST_SCROLL_BAR(priv->hscroll), hadjust);
-      g_signal_connect (hadjust, "changed", G_CALLBACK (
-                          child_adjustment_changed_cb), priv->hscroll);
-      child_adjustment_changed_cb (hadjust, priv->hscroll);
-    }
-}
-
-static void
-child_vadjustment_notify_cb (GObject    *gobject,
-                             GParamSpec *arg1,
-                             gpointer    user_data)
-{
-  StAdjustment *vadjust;
-
-  ClutterActor *actor = CLUTTER_ACTOR (gobject);
-  StScrollViewPrivate *priv = ST_SCROLL_VIEW (user_data)->priv;
-
-  vadjust = st_scroll_bar_get_adjustment (ST_SCROLL_BAR(priv->vscroll));
-  if (vadjust)
-    g_signal_handlers_disconnect_by_func (vadjust,
-                                          child_adjustment_changed_cb,
-                                          priv->vscroll);
-
-  st_scrollable_get_adjustments (ST_SCROLLABLE(actor), NULL, &vadjust);
-  if (vadjust)
-    {
-      /* Force scroll step if neede. */
-      if (priv->row_size_set)
-        {
-          g_object_set (vadjust,
-                        "step-increment", priv->row_size,
-                        NULL);
-        }
-
-      st_scroll_bar_set_adjustment (ST_SCROLL_BAR(priv->vscroll), vadjust);
-      g_signal_connect (vadjust, "changed", G_CALLBACK (
-                          child_adjustment_changed_cb), priv->vscroll);
-      child_adjustment_changed_cb (vadjust, priv->vscroll);
-    }
+  update_shadow_visibility (scroll);
 }
 
 static void
@@ -582,8 +883,24 @@ st_scroll_view_init (StScrollView *self)
 {
   StScrollViewPrivate *priv = self->priv = SCROLL_VIEW_PRIVATE (self);
 
-  priv->hscroll = CLUTTER_ACTOR (st_scroll_bar_new (NULL));
-  priv->vscroll = g_object_new (ST_TYPE_SCROLL_BAR, "vertical", TRUE, NULL);
+  priv->hscrollbar_policy = GTK_POLICY_AUTOMATIC;
+  priv->vscrollbar_policy = GTK_POLICY_AUTOMATIC;
+
+  priv->hadjustment = g_object_new (ST_TYPE_ADJUSTMENT, NULL);
+  priv->hscroll = g_object_new (ST_TYPE_SCROLL_BAR,
+                                "adjustment", priv->hadjustment,
+                                "vertical", FALSE,
+                                NULL);
+
+  priv->vadjustment = g_object_new (ST_TYPE_ADJUSTMENT, NULL);
+  g_signal_connect (priv->vadjustment, "changed",
+                    G_CALLBACK (child_adjustment_changed_cb), self);
+  g_signal_connect (priv->vadjustment, "notify::value",
+                    G_CALLBACK (child_adjustment_notify_value), self);
+  priv->vscroll = g_object_new (ST_TYPE_SCROLL_BAR,
+                                "adjustment", priv->vadjustment,
+                                "vertical", TRUE,
+                                NULL);
 
   clutter_actor_set_parent (priv->hscroll, CLUTTER_ACTOR (self));
   clutter_actor_set_parent (priv->vscroll, CLUTTER_ACTOR (self));
@@ -607,15 +924,8 @@ st_scroll_view_add (ClutterContainer *container,
       /* chain up to StBin::add() */
       st_scroll_view_parent_iface->add (container, actor);
 
-      /* Get adjustments for scroll-bars */
-      g_signal_connect (actor, "notify::hadjustment",
-                        G_CALLBACK (child_hadjustment_notify_cb),
-                        container);
-      g_signal_connect (actor, "notify::vadjustment",
-                        G_CALLBACK (child_vadjustment_notify_cb),
-                        container);
-      child_hadjustment_notify_cb (G_OBJECT (actor), NULL, container);
-      child_vadjustment_notify_cb (G_OBJECT (actor), NULL, container);
+      st_scrollable_set_adjustments (ST_SCROLLABLE (actor),
+                                     priv->hadjustment, priv->vadjustment);
     }
   else
     {
@@ -630,7 +940,8 @@ static void
 st_scroll_view_remove (ClutterContainer *container,
                        ClutterActor     *actor)
 {
-  StScrollViewPrivate *priv = ST_SCROLL_VIEW (container)->priv;
+  StScrollView *self = ST_SCROLL_VIEW (container);
+  StScrollViewPrivate *priv = self->priv;
 
   if (actor == priv->child)
     {
@@ -639,17 +950,23 @@ st_scroll_view_remove (ClutterContainer *container,
       /* chain up to StBin::remove() */
       st_scroll_view_parent_iface->remove (container, actor);
 
-      g_signal_handlers_disconnect_by_func (priv->child,
-                                            child_hadjustment_notify_cb,
-                                            container);
-      g_signal_handlers_disconnect_by_func (priv->child,
-                                            child_vadjustment_notify_cb,
-                                            container);
-      st_scrollable_set_adjustments ((StScrollable*) priv->child, NULL, NULL);
+      st_scrollable_set_adjustments (ST_SCROLLABLE (priv->child),
+                                     NULL, NULL);
 
       g_object_unref (priv->child);
       priv->child = NULL;
     }
+  else
+    {
+      if (actor == priv->vscroll)
+        priv->vscroll = NULL;
+      else if (actor == priv->hscroll)
+        priv->hscroll = NULL;
+      else
+        g_assert ("Unknown child removed from StScrollView");
+
+      clutter_actor_unparent (actor);
+    }
 }
 
 static void
@@ -667,6 +984,12 @@ st_scroll_view_foreach_with_internals (ClutterContainer *container,
 
   if (priv->vscroll != NULL)
     callback (priv->vscroll, user_data);
+
+  if (priv->top_shadow)
+    {
+      callback (priv->top_shadow, user_data);
+      callback (priv->bottom_shadow, user_data);
+    }
 }
 
 static void
@@ -724,14 +1047,11 @@ st_scroll_view_get_vscroll_bar (StScrollView *scroll)
 gfloat
 st_scroll_view_get_column_size (StScrollView *scroll)
 {
-  StAdjustment *adjustment;
   gdouble column_size;
 
   g_return_val_if_fail (scroll, 0);
 
-  adjustment = st_scroll_bar_get_adjustment (
-    ST_SCROLL_BAR (scroll->priv->hscroll));
-  g_object_get (adjustment,
+  g_object_get (scroll->priv->hadjustment,
                 "step-increment", &column_size,
                 NULL);
 
@@ -742,8 +1062,6 @@ void
 st_scroll_view_set_column_size (StScrollView *scroll,
                                 gfloat        column_size)
 {
-  StAdjustment *adjustment;
-
   g_return_if_fail (scroll);
 
   if (column_size < 0)
@@ -756,27 +1074,20 @@ st_scroll_view_set_column_size (StScrollView *scroll,
       scroll->priv->column_size_set = TRUE;
       scroll->priv->column_size = column_size;
 
-      adjustment = st_scroll_bar_get_adjustment (
-        ST_SCROLL_BAR (scroll->priv->hscroll));
-
-      if (adjustment)
-        g_object_set (adjustment,
-                      "step-increment", (gdouble) scroll->priv->column_size,
-                      NULL);
+      g_object_set (scroll->priv->hadjustment,
+                    "step-increment", (gdouble) scroll->priv->column_size,
+                    NULL);
     }
 }
 
 gfloat
 st_scroll_view_get_row_size (StScrollView *scroll)
 {
-  StAdjustment *adjustment;
   gdouble row_size;
 
   g_return_val_if_fail (scroll, 0);
 
-  adjustment = st_scroll_bar_get_adjustment (
-    ST_SCROLL_BAR (scroll->priv->vscroll));
-  g_object_get (adjustment,
+  g_object_get (scroll->priv->vadjustment,
                 "step-increment", &row_size,
                 NULL);
 
@@ -787,8 +1098,6 @@ void
 st_scroll_view_set_row_size (StScrollView *scroll,
                              gfloat        row_size)
 {
-  StAdjustment *adjustment;
-
   g_return_if_fail (scroll);
 
   if (row_size < 0)
@@ -801,13 +1110,9 @@ st_scroll_view_set_row_size (StScrollView *scroll,
       scroll->priv->row_size_set = TRUE;
       scroll->priv->row_size = row_size;
 
-      adjustment = st_scroll_bar_get_adjustment (
-        ST_SCROLL_BAR (scroll->priv->vscroll));
-
-      if (adjustment)
-        g_object_set (adjustment,
-                      "step-increment", (gdouble) scroll->priv->row_size,
-                      NULL);
+      g_object_set (scroll->priv->vadjustment,
+                    "step-increment", (gdouble) scroll->priv->row_size,
+                    NULL);
     }
 }
 
@@ -842,3 +1147,44 @@ st_scroll_view_get_mouse_scrolling (StScrollView *scroll)
 
   return priv->mouse_scroll;
 }
+
+/**
+ * st_scroll_view_set_policy:
+ * @scroll: A #StScrollView
+ * @hscroll: Whether to enable horizontal scrolling
+ * @vscroll: Whether to enable vertical scrolling
+ *
+ * Set the scroll policy.
+ */
+void
+st_scroll_view_set_policy (StScrollView   *scroll,
+                           GtkPolicyType   hscroll,
+                           GtkPolicyType   vscroll)
+{
+  StScrollViewPrivate *priv;
+
+  g_return_if_fail (ST_IS_SCROLL_VIEW (scroll));
+
+  priv = ST_SCROLL_VIEW (scroll)->priv;
+
+  if (priv->hscrollbar_policy == hscroll && priv->vscrollbar_policy == vscroll)
+    return;
+
+  g_object_freeze_notify ((GObject *) scroll);
+
+  if (priv->hscrollbar_policy != hscroll)
+    {
+      priv->hscrollbar_policy = hscroll;
+      g_object_notify ((GObject *) scroll, "hscrollbar-policy");
+    }
+
+  if (priv->vscrollbar_policy != vscroll)
+    {
+      priv->vscrollbar_policy = vscroll;
+      g_object_notify ((GObject *) scroll, "vscrollbar-policy");
+    }
+
+  clutter_actor_queue_relayout (CLUTTER_ACTOR (scroll));
+
+  g_object_thaw_notify ((GObject *) scroll);
+}
